@@ -12,9 +12,11 @@ from typing import (
     MutableMapping,
     NamedTuple,
     Optional,
+    Tuple,
     Union,
 )
 
+from mdformat import codepoints
 from mdformat.renderer._util import (
     CONSECUTIVE_KEY,
     RE_CHAR_REFERENCE,
@@ -41,6 +43,13 @@ if TYPE_CHECKING:
 
 LOGGER = logging.getLogger(__name__)
 
+# A marker used to point a location where word wrap is allowed
+# to occur.
+WRAP_POINT = "\x00"
+# A marker used to indicate location of a character that should be preserved
+# during word wrap. Should be converted to the actual character after wrap.
+PRESERVE_CHAR = "\x00"
+
 
 def make_render_children(separator: str) -> Render:
     def render_children(
@@ -82,6 +91,8 @@ def hardbreak(node: "RenderTreeNode", context: "RenderContext") -> str:
 
 
 def softbreak(node: "RenderTreeNode", context: "RenderContext") -> str:
+    if context.do_wrap and _in_paragraph(node):
+        return WRAP_POINT
     return "\n"
 
 
@@ -123,6 +134,9 @@ def text(node: "RenderTreeNode", context: "RenderContext") -> str:
     if text.endswith("!") and next_sibling and next_sibling.type == "link":
         text = text[:-1] + "\\!"
 
+    if context.do_wrap and _in_paragraph(node):
+        text = re.sub(r"\s+", WRAP_POINT, text)
+
     return text
 
 
@@ -214,6 +228,11 @@ def _render_inline_as_text(node: "RenderTreeNode", context: "RenderContext") ->
 
 def link(node: "RenderTreeNode", context: "RenderContext") -> str:
     text = "".join(child.render(context) for child in node.children)
+
+    if context.do_wrap:
+        # Prevent line breaks
+        text = text.replace(WRAP_POINT, " ")
+
     if node.info == "auto":
         return "<" + text + ">"
 
@@ -297,11 +316,16 @@ def _last_line_width(text: str) -> int:
     return width
 
 
-def _wrap(text: str, *, width: Union[int, Literal["no"]], preceding_text: str) -> str:
-    # Collapse all whitespace to a single space char
-    text = re.sub(r"\s+", " ", text)
+def _wrap(text: str, *, width: Union[int, Literal["no"]]) -> str:
+    """Wrap text at locations pointed by `WRAP_POINT`s.
+
+    Converts `WRAP_POINT`s to either a space or newline character, thus
+    wrapping the text. Already existing whitespace will be preserved as
+    is.
+    """
+    text, replacements = _prepare_wrap(text)
     if width == "no":
-        return text
+        return _recover_preserve_chars(text, replacements)
 
     wrapper = textwrap.TextWrapper(
         break_long_words=False,
@@ -309,26 +333,47 @@ def _wrap(text: str, *, width: Union[int, Literal["no"]], preceding_text: str) -
         width=width,
         expand_tabs=False,
         replace_whitespace=False,
-        drop_whitespace=False,
+    )
+    wrapped = wrapper.fill(text)
+    wrapped = _recover_preserve_chars(wrapped, replacements)
+    return " " + wrapped if text.startswith(" ") else wrapped
+
+
+def _prepare_wrap(text: str) -> Tuple[str, str]:
+    """Prepare text for wrap.
+
+    Convert `WRAP_POINT`s to spaces. Convert whitespace to
+    `PRESERVE_CHAR`s. Return a tuple with the prepared string, and
+    another string consisting of replacement characters for
+    `PRESERVE_CHAR`s.
+    """
+    result = ""
+    replacements = ""
+    for c in text:
+        if c == WRAP_POINT:
+            if not result or result[-1] != " ":
+                result += " "
+        elif c in codepoints.UNICODE_WHITESPACE:
+            result += PRESERVE_CHAR
+            replacements += c
+        else:
+            result += c
+    return result, replacements
+
+
+def _recover_preserve_chars(text: str, replacements: str) -> str:
+    replacement_iterator = iter(replacements)
+    return "".join(
+        next(replacement_iterator) if c == PRESERVE_CHAR else c for c in text
     )
 
-    # Prepend the text with as many null characters as the width of the last
-    # line in `preceding_text`. This forces a line break to happen sooner if
-    # the line already has content in `preceding_text`. Null characters are
-    # used because they can not naturally be present in the text.
-    text = _last_line_width(preceding_text) * "\x00" + text
 
-    # Do the wrapping
-    text = wrapper.fill(text)
-
-    # Remove the added null characters now that wrapping is done
-    text = text.lstrip("\x00")
-
-    # Because we set `drop_whitespace=False` for the wrapper, we now need
-    # to manually drop any whitespace surrounding a newline
-    text = re.sub(r"[\n ]*\n[\n ]*", "\n", text)
-
-    return text
+def _in_paragraph(node: "RenderTreeNode") -> bool:
+    while node.parent:
+        if node.parent.type == "paragraph":
+            return True
+        node = node.parent
+    return False
 
 
 def paragraph(node: "RenderTreeNode", context: "RenderContext") -> str:  # noqa: C901
@@ -336,29 +381,8 @@ def paragraph(node: "RenderTreeNode", context: "RenderContext") -> str:  # noqa:
 
     wrap_mode = context.options.get("mdformat", {}).get("wrap", "keep")
     if isinstance(wrap_mode, int) or wrap_mode == "no":
-        text = ""
-        buffer = ""
-        for child in inline_node.children:
-            if child.type in {"text", "softbreak"}:
-                buffer += child.render(context)
-            else:
-                if buffer:
-                    text += _wrap(buffer, width=wrap_mode, preceding_text=text)
-                buffer = ""
-
-                no_wrap_section = child.render(context)
-                # Add preceding wrap if the section extends
-                # beyond target wrap width
-                if (
-                    isinstance(wrap_mode, int)
-                    and text.endswith(" ")
-                    and _last_line_width(text) + _first_line_width(no_wrap_section)
-                    > wrap_mode
-                ):
-                    text = text[:-1] + "\n"
-                text += no_wrap_section
-        if buffer:
-            text += _wrap(buffer, width=wrap_mode, preceding_text=text)
+        text = "".join(child.render(context) for child in inline_node.children)
+        text = _wrap(text, width=wrap_mode)
     else:
         text = inline_node.render(context)
 
@@ -369,9 +393,8 @@ def paragraph(node: "RenderTreeNode", context: "RenderContext") -> str:  # noqa:
         if lines[i].startswith("\t"):
             lines[i] = "&#9;" + lines[i][1:]
 
-        # Make sure a paragraph line does not start with "#"
-        # (otherwise it will be interpreted as an ATX heading).
-        if lines[i].startswith("#"):
+        # If a line looks like an ATX heading, escape the first hash.
+        if re.match(r"#{1,6}( |\t|$)", lines[i]):
             lines[i] = f"\\{lines[i]}"
 
         # Make sure a paragraph line does not start with ">"
@@ -565,6 +588,11 @@ class RenderContext(NamedTuple):
     options: Mapping[str, Any]
     env: MutableMapping
 
+    @property
+    def do_wrap(self) -> bool:
+        wrap_mode = self.options.get("mdformat", {}).get("wrap", "keep")
+        return isinstance(wrap_mode, int) or wrap_mode == "no"
+
     def with_default_renderer_for(self, *syntax_names: str) -> "RenderContext":
         renderers = dict(self.renderers)
         for syntax in syntax_names:
