@@ -53,7 +53,7 @@ class SyntaxTreeNode:
             self._set_children_from_tokens(tokens)
             return
 
-        if not tokens:  # pragma: no cover
+        if not tokens:
             raise ValueError(
                 "Can only create root from empty token sequence."
                 " Set `create_root=True`."
